package org.apache.lucene.index.codecs.sep;

/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.apache.lucene.store.IndexInput;
import org.apache.lucene.store.Directory;
import org.apache.lucene.index.codecs.Codec;
import java.io.IOException;

/** Reads IndexInputs written with {@link
 * SingleIntIndexoutput} */
public class SingleIntIndexInput extends IntIndexInput {
  private final IndexInput in;

  public SingleIntIndexInput(Directory dir, String fileName, int readBufferSize)
    throws IOException {
    in = dir.openInput(fileName, readBufferSize);
    Codec.checkHeader(in, SingleIntIndexOutput.CODEC, SingleIntIndexOutput.VERSION_START);
  }

  public Reader reader() throws IOException {
    return new Reader((IndexInput) in.clone());
  }

  public void close() throws IOException {
    in.close();
  }

  public static class Reader extends IntIndexInput.Reader {
    // clone:
    private final IndexInput in;

    private final BulkReadResult result = new BulkReadResult();

    public Reader(IndexInput in) {
      this.in = in;
      result.offset = 0;
    }

    /** Reads next single int */
    public int next() throws IOException {
      return in.readVInt();
    }

    /** Reads next chunk of ints */
    public BulkReadResult read(int[] buffer, int count) throws IOException {
      result.buffer = buffer;
      for(int i=0;i<count;i++) {
        buffer[i] = in.readVInt();
      }
      result.len = count;
      return result;
    }

    public String descFilePointer() {
      return Long.toString(in.getFilePointer());
    }
  }

  private static class Index extends IntIndexInput.Index {
    private long fp;
    boolean first = true;

    public void read(IndexInput indexIn, boolean absolute)
      throws IOException {
      long cur = fp;
      if (absolute) {
        fp = indexIn.readVLong();
        first = false;
      } else {
        assert !first;
        fp += indexIn.readVLong();
      }
      if (Codec.DEBUG) {
        System.out.println("siii.idx.read: id=" + desc + " abs=" + absolute + " now=" + fp + " delta=" + (fp-cur));
      }
    }

    public void set(IntIndexInput.Index other) {
      fp = ((Index) other).fp;
      first = false;
    }

    public void seek(IntIndexInput.Reader other) throws IOException {
      ((Reader) other).in.seek(fp);
    }

    public String toString() {
      return Long.toString(fp);
    }
  }

  public Index index() {
    return new Index();
  }
}

