package org.apache.lucene.search;

/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.io.IOException;
import org.apache.lucene.index.TermRef;
import org.apache.lucene.index.TermsEnum;
import org.apache.lucene.index.DocsEnum;
import org.apache.lucene.util.Bits;

/**
 * Abstract class for enumerating a subset of all terms. 
 *
 * <p>On creation, the enumerator must already be position
 * to the first term.</p>
 * 
 * <p>Term enumerations are always ordered by
 * Term.compareTo().  Each term in the enumeration is
 * greater than all that precede it.</p>
*/
public abstract class FilteredTermsEnum extends TermsEnum {
    
  /** the delegate enum - to set this member use {@link #init} */
  protected TermsEnum actualEnum;
    
  /** Return true if term is acceptd */
  protected abstract boolean accept(TermRef term);
    
  /** Equality measure on the term */
  public abstract float difference();

  /** Return true if the enumeration should now stop */
  protected abstract boolean end();

  boolean done;

  protected abstract String field();

  protected TermRef current;
    
  /**
   * use this method to set the actual TermsEnum (e.g. in ctor),
   * it will be automatically positioned on the first matching term.
   */
  protected void init(TermsEnum actualEnum, TermRef term) throws IOException {
    this.actualEnum = actualEnum;
    done = false;

    // Find the first term that matches
    if (term != null) {
      SeekResult result = actualEnum.seek(term);

      final TermRef foundTerm;
      if (result.status == SeekResult.Status.FOUND) {
        foundTerm = term;
      } else if (result.status == SeekResult.Status.NOT_FOUND) {
        foundTerm = result.term;
      } else {
        foundTerm = null;
        done = true;
      }
      if (foundTerm != null) {
        if (accept(foundTerm)) {
          current = foundTerm;
        } else {
          next();
        }
      }
    }
  }

  // nocommit -- do we really need this?  caller can keep
  // track of it?
  protected TermRef term() {
    return current;
  }
    
  /** 
   * Returns the docFreq of the current Term in the enumeration.
   * Returns -1 if no Term matches or all terms have been enumerated.
   */
  public int docFreq() {
    assert actualEnum != null;
    return actualEnum.docFreq();
  }
    
  /** Increments the enumeration to the next element.  True if one exists. */
  public TermRef next() throws IOException {
    assert actualEnum != null;
    while (true) {
      if (done || end()) {
        current = null;
        return null;
      }
      TermRef term = actualEnum.next();
      if (term != null) {
        if (accept(term)) {
          current = term;
          return current;
        }
      } else {
        current = null;
        return null;
      }
    }
  }

  public SeekResult seek(TermRef term) throws IOException {
    SeekResult result = actualEnum.seek(term);
    if (result.status == SeekResult.Status.FOUND) {
      if (!accept(term)) {
        TermRef term2 = next();
        if (term2 == null) {
          result.status = SeekResult.Status.END;
        } else {
          result.status = SeekResult.Status.NOT_FOUND;
          result.term = term2;
        }
      } else {
        current = term;
      }
    } else if (result.status == SeekResult.Status.NOT_FOUND) {
      if (!accept(result.term)) {
        TermRef term2 = next();
        if (term2 == null) {
          result.status = SeekResult.Status.END;
        } else {
          result.term = term2;
        }
      } else {
        current = term;
      }
    }
    return result;
  }

  public DocsEnum docs(Bits bits) throws IOException {
    return actualEnum.docs(bits);
  }
}
