package org.apache.lucene.util.pfor;
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* When using the Sun 1.6 jvm, the performance tests below (using doDecompPerfTest... methods)
 * should be run with the -server argument to the forked jvm that is used for the
 * junit tests by adding this line just before the 1st batchtest line
 * in common-build.xml:
      <jvmarg value="-server"/>
 * Using this -server may be slow for other tests, in particular for shorter tests.
 */ 
 
import junit.framework.TestCase;
import java.nio.ByteBuffer;
import java.nio.IntBuffer;
import java.util.Arrays;
import org.apache.lucene.util.LuceneTestCase;


public class TestFrameOfRef extends LuceneTestCase {
  private static boolean doDecompPerfTests = true;

  private void showByte(int b, StringBuffer buf) { // for debugging
    for (int i = 7; i >= 0; i--) {
      buf.append((b >>> i) & 1);
    }
  }

  private void showBytes(byte[] array) { // for debugging
    StringBuffer buf = new StringBuffer();
    for (int i = 0; i < array.length; i++) {
      showByte(array[i] & 255, buf);
      if (((i+1) % 4) != 0) {
        buf.append(' ');
      } else {
        System.out.println(buf);
        buf.setLength(0);
      }
    }
  }

  /** Run compression without buffer, return the IntBuffer size needed for compression. */
  static int doNoBufferRun(int[] input, int offset, int numFrameBits, FrameOfRef frameOfRef) {
    frameOfRef.setUnCompressedData(input, offset, input.length - offset);
    frameOfRef.compress(numFrameBits);
    return frameOfRef.compressedSize();
  }

    FrameOfRef forNoBufferCompress = new FrameOfRef();

  /** Create an IntBuffer, compress the given input into this buffer, and return it. */
  static IntBuffer compressToBuffer(
        int[] input,
        int offset,
        int numFrameBits,
        int compressBufferSize,
        FrameOfRef frameOfRef) {
    // Allocate an IntBuffer as a view on a ByteBuffer
    ByteBuffer byteBuffer = ByteBuffer.allocate(4 * compressBufferSize);
    assert byteBuffer.hasArray();
    byte[] bufferByteArray = byteBuffer.array();
    assert bufferByteArray != null;
    IntBuffer compressBuffer = byteBuffer.asIntBuffer(); // no offsets used here.
    
    // Compress to buffer:
    frameOfRef.setUnCompressedData(input, offset, input.length - offset);
    frameOfRef.setCompressedBuffer(compressBuffer);
    frameOfRef.compress(numFrameBits);
// assert bufferByteArray.length == 4 * compressBufferSize; // for showBytes() below.
// showBytes(bufferByteArray);
    if (compressBufferSize >= 0) {
      //assertEquals("IntBuffer size after compress() to buffer", compressBufferSize, pforCompress.compressedSize());
    }
    return compressBuffer;
  }

  static void deCompressFromBufferVerify(
        IntBuffer compressBuffer,
        int[] input,
        int offset,
        int compressBufferSize,
        FrameOfRef frameOfRef) {
    // Decompress from the buffer:   
    frameOfRef.setCompressedBuffer(compressBuffer);
    assertEquals("Decompressed length before decompress()", input.length - offset, frameOfRef.decompressedSize());
    int[] output = new int[input.length]; // use same offset as input
    frameOfRef.setUnCompressedData(output, offset, output.length - offset);
    frameOfRef.decompress();
    assertEquals("IntBuffer size after decompress()", compressBufferSize, frameOfRef.compressedSize());
    if (! Arrays.equals(input, output)) {
      for (int i = 0; i < input.length; i++) {
        System.out.print("at index " + i + " output " + output[i]);
        System.out.print((input[i] != output[i]) ? " !=" : " ==");
        System.out.println(" input " + input[i]);
      }
      assertEquals("equal array lengths", input.length, output.length);
      assertTrue("input == output", Arrays.equals(input, output));
    }
  }

  private static void doTestOffset(
        int[] input,
        int offset,
        int numFrameBits,
        int compressBufferSize,
        String testName) {
System.out.println();
System.out.println(testName);
    FrameOfRef frameOfRef = new FrameOfRef();
    int actIntBufferSize = doNoBufferRun(input, offset, numFrameBits, frameOfRef);
    assertEquals("IntBuffer size after noBuffer run compress()", compressBufferSize, actIntBufferSize);
    FrameOfRef frameOfRef2 = new FrameOfRef();
    IntBuffer compressBuffer = compressToBuffer(input, offset, numFrameBits, actIntBufferSize, frameOfRef2);
    // Decompress and verify against original input.
    FrameOfRef frameOfRef3 = new FrameOfRef();
    deCompressFromBufferVerify(compressBuffer, input, offset, actIntBufferSize, frameOfRef3);
  }

  private void doTest(int[] input, int numBits, int compressBufferSize) {
    int offset = 0;
    doTestOffset(input, offset, numBits, compressBufferSize, getName());
  }

  public void test01NoExc() {
    int[] input = {1}; // no exception
    int numBits = 1;
    int compressBufferSize = 2;
    doTest(input, numBits, compressBufferSize);
  }

  public void test07_Offset1() {
    int[] input = {0,1};
    int offset = 1;
    int numBits = 1;
    int compressBufferSize = 2;
    doTestOffset(input, offset, numBits, compressBufferSize, getName());
  }

  public void test07_Offset2() {
    int[] input = new int[10];
    int offset = 9;
    input[offset] = 1;
    int numBits = 1;
    int compressBufferSize = 2;
    doTestOffset(input, offset, numBits, compressBufferSize, getName());
  }

  public void test12Series8Base4() {
    int[] input = {0,1,2,3,4,5,6,7,8};
    int numBits = 4;
    int compressBufferSize = 3;
    doTest(input, numBits, compressBufferSize);
  }

  public void test13Series8Base5() {
    int[] input = {0,1,2,3,4,5,6,7,8};
    int numBits = 5;
    int compressBufferSize = 3;
    doTest(input, numBits, compressBufferSize);
  }

  private void frameBitsForCompressionTest(int[] input, int expectedNumFrameBits) {
System.out.println();
System.out.println(getName());
    FrameOfRef frameOfRef = new FrameOfRef();
    frameOfRef.setUnCompressedData(input, 0, input.length);
    assertEquals("numFrameBits", expectedNumFrameBits, frameOfRef.frameBitsForCompression());
  }

  public void test20frameBitsForCompression() {
    int[] input = {2};
    frameBitsForCompressionTest(input, 2);
  }

  public void test21frameBitsForCompression() {
    int[] input = {9,8,7,6,5,4,3,2,1,0};
    frameBitsForCompressionTest(input, 4);
  }

  static void noBufferCompressionTest(int[] input, FrameOfRef frameOfRef, String testName) {
System.out.println();
System.out.println(testName);
    // Run compression without buffer:
    final int offset = 0;
    frameOfRef.setUnCompressedData(input, offset, input.length - offset);
    frameOfRef.compress();
System.out.println("Compress w/o buffer " + input.length + " ints into "
                    + frameOfRef.compressedSize()
                    + ", ratio " + (frameOfRef.compressedSize()/(float)input.length));
  }

  public void test30NoBufferCompression() {
    int[] input = {0,1,0,1,0,1,0,70000}; // would force exceptions for numFrameBits == 1
    noBufferCompressionTest(input, new FrameOfRef(), getName());
  }

  public void test31NoBufferCompression() {
    int[] input = {9,8,7,6,5,4,3,2,1,0,21,22,23,24,22,45,76,223,43,62,454};
    noBufferCompressionTest(input, new FrameOfRef(), getName());
  }

  public void test32NoBufferCompression() {
    int[] input = {9,8,7,6,5,4,3,2,1,0,21,22,23,24,22,45,76,223,43,62,454,
                   9,8,7,6,5,4,3,2,1,0,0};
    noBufferCompressionTest(input, new FrameOfRef(), getName());
  }

  public void test40For1Decompress() {
    int[] input = {
      1,0,1,0,1,0,1,0,
      1,0,1,0,1,0,1,0,
      1,0,1,0,1,0,1,0,
      1,0,1,0,1,0,1,0,
      1,0,1,0,1,0,1,0,
      1,0,1,0,1,0,1,0,
      1,0,1,0,1,0,1,0,
      1,0,1,0};
    int numBits = 1;
    int compressBufferSize = 3;
    doTest(input, numBits, compressBufferSize);
  }

  public void test41For2Decompress() {
    int[] input = {
      1,0,3,2,2,3,0,1,
      1,0,3,2,2,3,0,1,
      1,0,3,2,2,3,0,1,
      1,0,3,2};
    int numBits = 2;
    int compressBufferSize = 3;
    doTest(input, numBits, compressBufferSize);
  }

  public void test42For3Decompress() {
    int[] input = {
      1,0,3,2,7,6,5,4,
      7,5,4,5,6,7,0,1,
      1,0,3,6,4,7,5,1,
      1,0,4,5,6,7,0,1, // 32 input, 3 ints compressed
      1,0,4,5,6,7,0,1,
      4,6,3,6,4,7,5,1,
      1,0,4,5,6,7}; // 22 more input, 9 bytes compressed
    int numBits = 3;
    int compressBufferSize = 7;
    doTest(input, numBits, compressBufferSize);
  }
  
  public void test43For4Decompress() {
    int[] input = {
      1,0,3,2,5,7,4,6,
      8,9,10,2,15,0};
    int numBits = 4;
    int compressBufferSize = 3;
    doTest(input, numBits, compressBufferSize);
  }

  public void test447For17Decompress() {
    int[] input = {1,1022,1023};
    int numBits = 17;
    int compressBufferSize = 3;
    doTest(input, numBits, compressBufferSize);
  }

  public void test449For17Decompress() {
    int[] input = {
      1,1022,1023,127,126,13,32768,32767,
      16383,16382,35,37,63,2046,2047,60,
      9,4094,4095,14,511,510,13,9,
      23,21,8190,8191,226,255,65536,65535};
    int numBits = 17;
    int compressBufferSize = 18;
    doTest(input, numBits, compressBufferSize);
  }

  /** Repeat decompressing from a given buffer, report on performance. */
  static void doDecompPerfTestUsing(
        IntBuffer comprDataBuffer,
        int decomprSize,
        FrameOfRef frameOfRef,
        String testName) {
    frameOfRef.setCompressedBuffer(comprDataBuffer);
    for (int rep = 0; rep < 3; rep++) {
      long maxTestMillis = 300; // 1000
      long testMillis;
      int iterations = 0;
      long startMillis = System.currentTimeMillis();
      final int decompsPerIter = 1024 * 128;
      int[] output = new int[decomprSize]; // use 0 offset
      frameOfRef.setUnCompressedData(output, 0, decomprSize); // FIXME: use decomprSize from compressBuffer, or even from pforDecompress
      do {
        for (int i = 0; i < decompsPerIter; i++) {
          frameOfRef.decompress();
        }
        iterations++;
        testMillis = System.currentTimeMillis() - startMillis;
      } while ((testMillis < maxTestMillis) && (iterations < 1000));
      long totalDecompressed = (((long) decomprSize) * decompsPerIter * iterations);
  System.out.println(testName + " " + rep
      + " bits " + frameOfRef.getNumFrameBits()
      + " decompressed " + totalDecompressed
      + " in " + testMillis + " msecs, "
      + ((int)(totalDecompressed/(testMillis * 1000f))) + " kints/msec, ("
      + iterations + " iters).");
    }
  }
  

  private void doDecompPerfTest(IntBuffer comprDataBuffer, int decomprSize) {
    if (! doDecompPerfTests) {
      return;
    }
    FrameOfRef frameOfRef = new FrameOfRef();
    doDecompPerfTestUsing(comprDataBuffer, decomprSize, frameOfRef, getName());
  }

  private void doCompDecompPerfTest(int[] decomprData, int numBits) {
    int offset = 0;
    FrameOfRef frameOfRef = new FrameOfRef();
    int actIntBufferSize = doNoBufferRun(decomprData, offset, numBits, frameOfRef);
    FrameOfRef frameOfRef2 = new FrameOfRef();
    IntBuffer comprDataBuffer = compressToBuffer(decomprData, offset, numBits, actIntBufferSize, frameOfRef2);
    if (! doDecompPerfTests) {
      return;
    }
    FrameOfRef frameOfRef3 = new FrameOfRef();
    doDecompPerfTestUsing(comprDataBuffer, decomprData.length, frameOfRef3, getName());
  }

  public void test91PerfFor01Decompress() {
    int[] input = {
      1,0,1,0,1,0,1,0,
      1,0,1,0,1,0,1,0,
      1,0,1,0,1,0,1,0,
      1,0,1,0,1,0,1,0};
    int numBits = 1;
    doCompDecompPerfTest(input, numBits);
  }

  public void test91PerfFor02Decompress() {
    int[] input = {
      1,0,3,2,3,2,1,0,
      1,0,3,2,3,2,1,0,
      1,0,3,2,3,1,0,1,
      1,0,3,2,3,2,1,0};
    int numBits = 2;
    doCompDecompPerfTest(input, numBits);
  }

  public void test91PerfFor03Decompress() {
    int[] input = {
      1,0,3,2,7,6,5,4,
      7,5,4,5,6,7,0,1,
      1,0,3,6,4,7,5,1,
      1,0,4,5,6,7,0,1};
    int numBits = 3;
    doCompDecompPerfTest(input, numBits);
  }

  public void test91PerfFor04Decompress() {
    int[] input = {
      1,0,3,2,7,6,5,4,
      9,8,11,14,12,15,13,9,
      7,5,4,5,6,7,0,1,
      9,8,11,14,12,15,13,9};
    int numBits = 4;
    doCompDecompPerfTest(input, numBits);
  }

  public void test91PerfFor05Decompress() {
    int[] input = {
      1,0,3,2,7,6,5,4,
      9,8,11,14,12,15,13,9,
      23,21,20,21,22,23,16,17,
      9,8,11,14,12,15,13,9};
    int numBits = 5;
    doCompDecompPerfTest(input, numBits);
  }

  public void test91PerfFor06Decompress() {
    int[] input = {
      1,0,3,2,7,6,5,4,
      33,32,35,37,63,62,61,60,
      9,8,11,14,12,15,13,9,
      23,21,20,21,22,23,16,17};
    int numBits = 6;
    doCompDecompPerfTest(input, numBits);
  }

  public void test91PerfFor07Decompress() {
    int[] input = {
      1,0,3,2,7,6,5,4,
      33,32,35,37,63,62,61,60,
      9,8,11,14,127,126,13,9,
      23,21,20,21,22,23,16,17};
    int numBits = 7;
    doCompDecompPerfTest(input, numBits);
  }

  public void test91PerfFor08Decompress() {
    int[] input = {
      1,0,3,127,126,13,4,7,
      33,32,35,37,63,62,61,60,
      9,8,11,14,127,126,13,9,
      23,21,20,21,226,255,16,17};
    int numBits = 8;
    doCompDecompPerfTest(input, numBits);
  }

  public void test91PerfFor09Decompress() {
    int[] input = {
      1,0,3,127,126,13,4,7,
      33,32,35,37,63,62,61,60,
      9,8,11,14,511,510,13,9,
      23,21,20,21,226,255,16,17};
    int numBits = 9;
    doCompDecompPerfTest(input, numBits);
  }

  public void test91PerfFor10Decompress() {
    int[] input = {
      1,1022,1023,127,126,13,4,7,
      33,32,35,37,63,62,61,60,
      9,8,11,14,511,510,13,9,
      23,21,20,21,226,255,16,17};
    int numBits = 10;
    doCompDecompPerfTest(input, numBits);
  }

  public void test91PerfFor11Decompress() {
    int[] input = {
      1,1022,1023,127,126,13,4,7,
      33,32,35,37,63,2046,2047,60,
      9,8,11,14,511,510,13,9,
      23,21,20,21,226,255,16,17};
    int numBits = 11;
    doCompDecompPerfTest(input, numBits);
  }

  public void test91PerfFor12Decompress() {
    int[] input = {
      1,1022,1023,127,126,13,4,7,
      33,32,35,37,63,2046,2047,60,
      9,4094,4095,14,511,510,13,9,
      23,21,20,21,226,255,16,17};
    int numBits = 12;
    doCompDecompPerfTest(input, numBits);
  }

  public void test91PerfFor13Decompress() {
    int[] input = {
      1,1022,1023,127,126,13,4,7,
      33,32,35,37,63,2046,2047,60,
      9,4094,4095,14,511,510,13,9,
      23,21,8190,8191,226,255,16,17};
    int numBits = 13;
    doCompDecompPerfTest(input, numBits);
  }

  public void test91PerfFor14Decompress() {
    int[] input = {
      1,1022,1023,127,126,13,4,7,
      16383,16382,35,37,63,2046,2047,60,
      9,4094,4095,14,511,510,13,9,
      23,21,8190,8191,226,255,16,17};
    int numBits = 14;
    doCompDecompPerfTest(input, numBits);
  }

  public void test91PerfFor15Decompress() {
    int[] input = {
      1,1022,1023,127,126,13,32766,32767,
      16383,16382,35,37,63,2046,2047,60,
      9,4094,4095,14,511,510,13,9,
      23,21,8190,8191,226,255,16,17};
    int numBits = 15;
    doCompDecompPerfTest(input, numBits);
  }

  public void test91PerfFor16Decompress() {
    int[] input = {
      1,1022,1023,127,126,13,32768,32767,
      16383,16382,35,37,63,2046,2047,60,
      9,4094,4095,14,511,510,13,9,
      23,21,8190,8191,226,255,65534,65535};
    int numBits = 16;
    doCompDecompPerfTest(input, numBits);
  }

  public void test91PerfFor17Decompress() {
    int[] input = {
      1,1022,1023,127,126,13,32768,32767,
      16383,16382,35,37,63,2046,2047,60,
      9,4094,4095,14,511,510,13,9,
      23,21,8190,8191,226,255,65536,65535};
    int numBits = 17;
    doCompDecompPerfTest(input, numBits);
  }

  public void test91PerfFor18Decompress() {
    int[] input = {
      1,1022,1023,127,126,13,32768,32767,
      16383,16382,35,37,63,2046,2047,60,
      9,4094,4095,14,511,131071,131072,9,
      23,21,8190,8191,226,255,65536,65535};
    int numBits = 18;
    doCompDecompPerfTest(input, numBits);
  }

  public void test91PerfFor19Decompress() {
    int[] input = {
      1,1022,1023,127,262144,262143,4,7,
      16383,16382,35,37,63,2046,2047,60,
      9,4094,4095,14,511,131071,131072,9,
      23,21,8190,8191,226,255,16,17};
    int numBits = 19;
    doCompDecompPerfTest(input, numBits);
  }

  public void test91PerfFor20_32Decompress() {
    for (int numBits = 20; numBits <= 32; numBits++) {
      int[] input = {
        1,(int)((1L<<numBits)-2),(int)((1L<<numBits)-1),127,262144,262143,4,7,
        16383,16382,35,37,63,2046,2047,60,
        9,4094,4095,14,511,131071,131072,9,
        23,21,8190,8191,226,255,0,17};
      doCompDecompPerfTest(input, numBits);
    }
  }
}
