package org.apache.lucene.util.pfor;
import org.apache.lucene.store.*;
import org.apache.lucene.util.LuceneTestCase;
import junit.framework.TestCase;
import java.io.IOException;

import java.nio.ByteBuffer;
import java.nio.IntBuffer;
import java.util.Random;
import java.text.NumberFormat;

public class TestPFor2 extends LuceneTestCase {

  private static final int BLOCK_SIZE = 128;

  public static void main(String[] args) throws Throwable {
    Directory dir = FSDirectory.getDirectory(args[0]);

    if (args.length != 3) {
      System.out.println("\nUsage: java org.apache.lucene.util.TestPFor2 <indexDirName> <vIntFileNameIn> <pForFileNameOut>\n");
      System.out.println("Eg: java org.apache.lucene.util.TestPFor2 /lucene/index _l.prx _l.prx.pfor\n");
      System.exit(1);
    }

    String vIntFileNameIn = args[1];
    String pForFileNameOut = args[2];

    // Convert vInt encoding --> pfor
    if (!dir.fileExists(pForFileNameOut)) {
      System.out.println("\nencode " + vIntFileNameIn + " to " + pForFileNameOut + "...");
      convertVIntToPFor(dir, vIntFileNameIn, pForFileNameOut);
    }

    System.out.println("\ndecompress using pfor:");
    long bestPFor = 0;
    for(int round=0;round<5;round++) {
      long speed = readPFor(dir, pForFileNameOut);
      if (speed > bestPFor)
        bestPFor = speed;
    }

    System.out.println("\ndecompress using readVInt:");
    long bestVInt = 0;
    for(int round=0;round<5;round++) {
      long speed = readVInts(dir, vIntFileNameIn);
      if (speed > bestVInt)
        bestVInt = speed;
    }

    NumberFormat nf = NumberFormat.getInstance();
    if (bestVInt > bestPFor)
      System.out.println("\nPFor is " + nf.format((bestVInt-bestPFor)*100.0/bestVInt) + "% slower");
    else
      System.out.println("\nPFor is " + nf.format((bestPFor-bestVInt)*100.0/bestVInt) + "% faster");

    dir.close();
  }

  /** Returns ints/sec speed */
  public static long readVInts(Directory dir, String vIntFileNameIn) throws Throwable {
    IndexInput in = dir.openInput(vIntFileNameIn);
    final long t0 = System.currentTimeMillis();
    long count = 0;
    while(true) {
      try {
        in.readVInt();
        count++;
      } catch (IOException ioe) {
        break;
      }
    }
    final long t1 = System.currentTimeMillis();
    in.close();
    System.out.println((t1-t0) + " msec to read " + count + " ints (" + (count/(t1-t0)) + " ints/msec)");

    return count/(t1-t0);
  }

  /** Returns ints/sec speed */
  public static long readPFor(Directory dir, String pForFileNameOut) throws Throwable {
    IndexInput in = dir.openInput(pForFileNameOut);

    PFor pforDecompress = new PFor();
    ByteBuffer byteBuffer = ByteBuffer.allocate(1024);
    byte[] bufferByteArray = byteBuffer.array();
    IntBuffer compressBuffer = byteBuffer.asIntBuffer(); // no offsets used here.
    pforDecompress.setCompressedBuffer(compressBuffer);
    final int[] temp = new int[BLOCK_SIZE];
    final long t0 = System.currentTimeMillis();
    long count = 0;
    while(true) {
      try {
        int numByte = in.readInt();
        in.readBytes(bufferByteArray, 0, numByte);
        pforDecompress.setUnCompressedData(temp, 0, 0);
        pforDecompress.decompress();
        count++;
      } catch (IOException ioe) {
        break;
      }
    }
    final long t1 = System.currentTimeMillis();
    System.out.println((t1-t0) + " msec to decode " + (BLOCK_SIZE*count) + " ints (" + (BLOCK_SIZE*count/(t1-t0)) + " ints/msec)");
    in.close();

    return (BLOCK_SIZE*count)/(t1-t0);
  }

  // nocommit -- so JUnit doesn't complain
  public void test() {
    assertTrue(true);
  }

  public static void convertVIntToPFor(Directory dir, String vIntFileNameIn, String pForFileNameOut) throws Throwable {
    IndexInput in = dir.openInput(vIntFileNameIn);
    IndexOutput out = dir.createOutput(pForFileNameOut);

    ByteBuffer byteBuffer = ByteBuffer.allocate(1024);
    byte[] bufferByteArray = byteBuffer.array();
    IntBuffer compressBuffer = byteBuffer.asIntBuffer(); // no offsets used here.

    PFor pforCompress = new PFor();
    pforCompress.setCompressedBuffer(compressBuffer);

    // Get ints
    int count = 0;
    int upto = 0;
    int[] temp = new int[BLOCK_SIZE];

    final Random r = new Random();
    final int[] counts = new int[32];

    while(true) {
      try {
        temp[upto++] = in.readVInt();
      } catch (IOException ioe) {
        break;
      }
      if (upto == BLOCK_SIZE) {
        pforCompress.setUnCompressedData(temp, 0, BLOCK_SIZE);
        final int numFrameBits = pforCompress.frameBitsForCompression();
        counts[numFrameBits]++;
        pforCompress.compress();
        final int numByte = pforCompress.compressedSize() * 4;
        out.writeInt(numByte);
        out.writeBytes(bufferByteArray, 0, numByte);
        upto = 0;
        count++;
      }
    }
    in.close();
    out.close();
    System.out.println((BLOCK_SIZE*count) + " ints; " + dir.fileLength(pForFileNameOut) + " bytes compressed vs orig size " + dir.fileLength(vIntFileNameIn));
 
    /*
    NumberFormat nf = NumberFormat.getInstance();
    for(int i=1;i<31;i++)
      System.out.println(i + " bits: " + counts[i] + " [" + nf.format(100.0*counts[i]/count) + " %]");
    */
  }
}